/**
 * Provide Base64 encoding.
 *
 * @author Simon Parsons
 *
 * Represent raw byte data as ASCII characters, representing 6 bits
 * with each ASCII digit.
 *
 * Based on Knudsen, "Java Cryptography", pages 258-262.
 */
public class Base64{

    // Base64 is a class that supports conversion between raw bit
    // streams and ASCII. We interpret the bit stream as a sequence of
    // base 64 numbers (each of which is 6 bits). Four such numbers
    // are then equivalent to three ASCII chracters.
    //
    // The conversion allows raw bits to be transmitted by mechanisms 
    // that expect character data, like mailers and print routines.
    //
    // There is a standard for this conversion, and Knuden's original 
    // code met it. I made some changes for (what I thought was) clarity
    // and hopefully we are still doing things the standard way.

    // Encode turns a bit stream (byte array) into a String (sequence of
    // ASCII characters). We use it to make the output of some encryption
    // a sequence of characters that we can read.
    public static String encode(byte[] raw){
	StringBuffer encoded = new StringBuffer();
	for(int i = 0; i < raw.length; i += 3){
	    encoded.append(encodeBlock(raw, i));
	}
	return encoded.toString();
    }

    // Helper class for the above. Expects to be fed 3 bytes of 
    // input data and uses it to generate 4 base 64 digits.
    protected static char[] encodeBlock(byte[] raw, int offset){
	int block = 0;
	// slack tells us how much space is left in the input 
	// byte array
	int slack = raw.length - offset -1;
	// end is then the number of input bytes we can read ---
	// this could be less than 3 if we are at the end.
	int end;
	if(slack >= 2){
	    end = 2;
	}
	else{
	    end = slack;
	}
        
	// Now loop through the input data, converting it to unsigned
	// form, shifting so it slots in the right place, and adding it 
	// to block.
	for(int i = 0; i <= end; i++){
	    byte b = raw[offset + i];
	    int neuter;
	    if(b < 0){
		neuter = b + 256;
	    }
	    else{
		neuter = b;
		}
	    block += neuter << (8 * (2 - i));
	}

	// Now extract the base64 digits from block. Each of these is
	// a six bit quantity. Again we have to do some suitable 
	// shifting.
	char[] base64 = new char[4];
	for(int i = 0; i < 4; i++){
	    int sixbit = (block >>> (6 * (3 - i))) & 0x3f;
	    base64[i] = getChar(sixbit);
	}

	// Add padding to make up the slack.
	if(slack < 1){
	    base64[2] = '=';
	}
	if(slack < 2){
	    base64[3] = '=';
	}

	return base64;
    }

    // This is how we translate from a six bit quantity to a base64 
    // digit.
    protected static char getChar(int sixBit){
	if(sixBit >= 0 && sixBit <= 25){
	    return (char)('A' + sixBit);
	}
	else
	if(sixBit >= 26 && sixBit <= 51){
	    return (char)('a' + (sixBit - 26));
	}
	else
	if(sixBit >= 52 && sixBit <= 61){
	    return (char)('0' + (sixBit - 52));
	}
	else
	if(sixBit == 62){
	    return '+';
	}
	else
	if(sixBit == 63){
	    return '/';
	}
	else return '?';
    }

    // Decode takes a sequence of ASCII characters and constructs a sequence of
    // bits (byte array) from it.
    public static byte[] decode(String base64){
	// Figure out how many padding digits there are.
	int pad = 0;
	for(int i= base64.length() - 1; base64.charAt(i) == '='; i--){
	    pad++;
	}

	// Calculate the lenth of the byte array
	int length = base64.length() * 6 / 8 - pad;
	byte[] raw = new byte[length];

	// Now go through the base 64 string, pulling out four digits
	// and placing each one into a byte.
	int rawIndex = 0;
	for(int i = 0; i < base64.length(); i += 4){
	    int block = (getValue(base64.charAt(i)) << 18)
		+ (getValue(base64.charAt(i + 1)) << 12)
		+ (getValue(base64.charAt(i + 2)) << 6)
		+ (getValue(base64.charAt(i + 3)));

	    for(int j = 0; j < 3 && rawIndex + j < raw.length; j++){
		raw[rawIndex + j] = (byte)((block >> (8 * (2 - j))) & 0xff);
	    }
	    rawIndex +=3;
	}
	return raw;
    }

    // Helper method for decode. Map from a byte's worth of data to the
    // six-bit equivalent.
    protected static int getValue(char c){
	if(c >= 'A' && c <= 'Z'){
	    return c - 'A';
	}
	else
	if(c >= 'a' && c <= 'z'){
	    return c - 'a' + 26;
	}
	else
	if(c >= '0' && c <= '9'){
	    return c - '0' + 52;
	}
	else
	if(c == '+'){
	    return 62;
	}
	else
	if(c == '/'){
	    return 63;
	}
	else
	if(c == '='){
	    return 0;
	}
	else{
	    return -1;
	}
    }
}